import math
import cv2
import numpy as np


def _headTrackingHeadingCalculation(self, i, blur, thresh1, thresh2, frameOri, erodeSize, frame_width, frame_height, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingProbabilityOfGoodDetection, headPosition, lengthX, xmin=0, ymin=0, wellNumber=-1, oldFrameList=[]):
  xHB_TN = 0
  heading = 0
  x = 0
  y = 0
  lastFirstTheta = 0

  if self._hyperparameters["fixedHeadPositionX"] != -1:

    trackingHeadTailAllAnimals[0, i-self._firstFrame][0][0] = int(self._hyperparameters["fixedHeadPositionX"])
    trackingHeadTailAllAnimals[0, i-self._firstFrame][0][1] = int(self._hyperparameters["fixedHeadPositionY"])

  else:

    if (self._hyperparameters["headEmbeded"] == 1 and i == self._firstFrame) or (self._hyperparameters["headEmbeded"] == 0) or (self._hyperparameters["headEmbededTeresaNicolson"] == 1):

      if self._hyperparameters["findHeadPositionByUserInput"] == 0:

        # Finds head position for frame i
        takeTheHeadClosestToTheCenter = self._hyperparameters["takeTheHeadClosestToTheCenter"]
        
        (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(blur)
        
        if "localMinimumDarkestThreshold" in self._hyperparameters and self._hyperparameters["localMinimumDarkestThreshold"]:
          localMinimumDarkestThreshold = int(self._hyperparameters["localMinimumDarkestThreshold"])
        else:
          localMinimumDarkestThreshold = 180
        
        for animalNumber in range(self._hyperparameters["nbAnimalsPerWell"]):
          
          if minVal < localMinimumDarkestThreshold:
            
            if type(trackingProbabilityOfGoodDetection) != int and len(trackingProbabilityOfGoodDetection) and i-self._firstFrame < len(trackingProbabilityOfGoodDetection[0]):
              trackingProbabilityOfGoodDetection[0, i-self._firstFrame] = np.sum(255 - blur)

            x = headPosition[0]
            y = headPosition[1]

            if type(headPosition) == tuple:
              headPosition = list(headPosition)
              headPosition[0] = headPosition[0] + xmin
              headPosition[1] = headPosition[1] + ymin
              headPosition = tuple(headPosition)
            else:
              headPosition[0] = headPosition[0] + xmin
              headPosition[1] = headPosition[1] + ymin

            # Calculate heading for frame i
            if type(thresh1) != int:
              [heading, lastFirstTheta] = self._calculateHeading(x, y, i, thresh1, thresh2, takeTheHeadClosestToTheCenter, 0, wellNumber)
            
            if "removeShades" in self._hyperparameters and self._hyperparameters["removeShades"]:
              okAll = True
              if i > 110:
                for oldFrame in oldFrameList:
                  compareDarknestWithThePastWindow = int(self._hyperparameters["headSize"] / 40)
                  centeredROIValue = np.mean(frameOri[headPosition[1]-compareDarknestWithThePastWindow:headPosition[1]+compareDarknestWithThePastWindow, headPosition[0]-compareDarknestWithThePastWindow:headPosition[0]+compareDarknestWithThePastWindow])
                  centeredROIValueOld = np.mean(oldFrame[headPosition[1]-compareDarknestWithThePastWindow:headPosition[1]+compareDarknestWithThePastWindow, headPosition[0]-compareDarknestWithThePastWindow:headPosition[0]+compareDarknestWithThePastWindow])
                  if self._hyperparameters["brightAnimalDarkBackround"]:
                    ok = (centeredROIValue > centeredROIValueOld)
                  else:
                    ok = (centeredROIValue < centeredROIValueOld)
                  okAll = (okAll and ok)
            
            if minVal < localMinimumDarkestThreshold and okAll:
              trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][0] = headPosition[0]
              trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][1] = headPosition[1]
              trackingHeadingAllAnimals[animalNumber, i-self._firstFrame] = heading
            else:
              trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][0] = 0
              trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][1] = 0
              trackingHeadingAllAnimals[animalNumber, i-self._firstFrame] = 0
            
            cv2.circle(blur, headPosition, self._hyperparameters["headSize"], (255, 255, 255), -1)
            (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(blur)

      else:

        # This is for the head-embedeed: at this point, this is set again in the tail tracking (the heading is set in the tail tracking as well)
        trackingHeadTailAllAnimals[0, i-self._firstFrame][0][0] = headPosition[0]
        trackingHeadTailAllAnimals[0, i-self._firstFrame][0][1] = headPosition[1]

    else:
      # If head embeded, heading and head position stay the same for all frames
      trackingHeadingAllAnimals[0, i-self._firstFrame]  = trackingHeadingAllAnimals[0, 0]
      trackingHeadTailAllAnimals[0, i-self._firstFrame] = trackingHeadTailAllAnimals[0, 0]

  return lastFirstTheta



from .createDistanceBetweenFramesExcelFile import *
from .getCurvaturePerBout import *
from .getCurvaturePerTimeInterval import *
from .getDataPerBout import *
from .getDataPerTimeInterval import *
from .getFirstAndLastFrame import *
from .getKinematicParametersPerBout import *
from .getKinematicParametersPerInterval import *
from .getTailAngleHeatmapPerBout import *
from .getTailAngleHeatmapPerTimeInterval import *
from .listAllBouts import *
from .plotCurvatureYaxisExact import *
from .plotCurvatureYaxisApproximate import *
from .plotTailAngleHeatmap import *
from .setFPSandPixelSize import *
from .getFPSandPixelSize import *
from .applyMedianFilterOnCurvature import *
from .getConfigurationFileUsed import *
from .getNbOscAndTBFPerBoutFromManualClassification import *
from .plotManualVsAutomaticBendLocations import *
from .plotKinematicParametersHist import *
from .deleteAutomaticallyCalculatedKinematicParameters import *
from .createExcelFileWithRawData import *
from .plotSingleFrameTrackingPoints import *
from .getSingleFrameTrackingAndCurvatureForFolder import *
from .plotSingleFrameTrackingPoints import *
from .getSingleFrameDistanceToAxis import *
from .getSingleFrameTrackingAndDistanceToAxisForFolder import *
from .setDataPerTimeInterval import *
from .getWellCoordinates import *
from .getNumberOfAnimalsPerWell import *
from .getNumberOfWells import *
from .reassignMultipleAnimalsId import reassignMultipleAnimalsId
from .detectBouts import *
from .smoothHeadPositions import *



import numpy as np

from ._openResultsFile import openResultsFile


def setDataPerTimeInterval(videoName: str, numWell: int, numAnimal: int, startTimeInSeconds: float, endTimeInSeconds: float, parameterName: str, newValues: np.array):
  if endTimeInSeconds is not None and startTimeInSeconds is not None and startTimeInSeconds >= endTimeInSeconds:
    raise ValueError('end time must be larger than start time')
  with openResultsFile(videoName, 'a') as results:
    if not startTimeInSeconds and endTimeInSeconds is None:
      intervalStart = 0
      intervalEnd = results.attrs['lastFrame'] - results.attrs['firstFrame']
    else:
      if 'videoFPS' not in results.attrs:
        raise ValueError(f'videoFPS not found in the results, cannot convert seconds to frames')
      firstFrame = results.attrs['firstFrame']
      firstFrameInSeconds = firstFrame / results.attrs['videoFPS']
      if startTimeInSeconds == 0 and firstFrame == 1:
        # when running tracking on the whole video, firstFrame is 1, but from the user's perspective, startTimeInSeconds is 0
        startTimeInSeconds = firstFrameInSeconds
      lastFrameInSeconds = results.attrs['lastFrame'] / results.attrs['videoFPS']
      if startTimeInSeconds is None:
        startTimeInSeconds = firstFrameInSeconds
      if endTimeInSeconds is None:
        endTimeInSeconds = lastFrameInSeconds
      if startTimeInSeconds < firstFrameInSeconds or endTimeInSeconds > lastFrameInSeconds:
        raise ValueError(f'Tracking was performed from {firstFrameInSeconds}s to {lastFrameInSeconds}s, start and end times must be within this interval.')
      intervalStart = int(startTimeInSeconds * results.attrs['videoFPS']) - firstFrame
      intervalEnd = int(endTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    if parameterName not in dataGroup:
      allParameters = ', '.join(dataGroup.keys())
      raise ValueError(f'Parameter "{parameterName}" not found in results. Available parameters: {allParameters}')
    dataset = dataGroup[parameterName]
    if parameterName == 'HeadPos':
      dataset[intervalStart:intervalEnd] = [(newValues[i, 0], newValues[i, 1]) for i in range(len(newValues))]
    else:
      dataset[intervalStart:intervalEnd] = newValues



import numpy as np
from scipy.interpolate import UnivariateSpline

#Heading change in degrees
def getDeltaHead(b, fromHeadingIdx=None):
  if fromHeadingIdx is not None:
    if 'Heading' not in b:
      return np.nan
    slope0 = b['Heading'][0] * 180 / np.pi
    slope1 = b['Heading'][fromHeadingIdx] * 180 / np.pi
  else:
    numps = int(len(b["TailX_VideoReferential"][0]) / 3) # Taking point at a third of the lenght of the tail
    if numps == 0:
      numps = 1
    bXs = np.concatenate((b["TailX_VideoReferential"][0][-numps:],[b["HeadX"][0]]))
    bYs = np.concatenate((b["TailY_VideoReferential"][0][-numps:],[b["HeadY"][0]]))
    slope0 = np.arctan2((bYs[-1]-bYs[0]),(bXs[-1]-bXs[0]))*180/np.pi
    bXs = np.concatenate((b["TailX_VideoReferential"][-1][-numps:],[b["HeadX"][-1]]))
    bYs = np.concatenate((b["TailY_VideoReferential"][-1][-numps:],[b["HeadY"][-1]]))
    slope1 = np.arctan2((bYs[-1]-bYs[0]),(bXs[-1]-bXs[0]))*180/np.pi
  
  delt = slope1 - slope0
  if delt > 180:
    return -(360 - delt)
  elif delt < -180:
    return 360 + delt
  else:
    return delt


import os
import subprocess

import cv2
import ffmpeg_progress_yield
import static_ffmpeg
static_ffmpeg.add_paths()

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QFileDialog, QProgressDialog

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import zebrazoom.code.util as util


def _runLegacy(inputPath, outputPath, progressDialog, firstFrame, lastFrame):
    cap = zzVideoReading.VideoCapture(inputPath)
    if not cap.isOpened():
      print("Error opening video stream or file")
      return
    cap.set(1, firstFrame)
    width = int(cap.get(3))
    height = int(cap.get(4))

    out = cv2.VideoWriter(outputPath, cv2.VideoWriter_fourcc(*'MJPG'), 10, (width, height))
    for frameIdx in range(firstFrame, lastFrame):
      progressDialog.setValue(frameIdx)
      progressDialog.setLabelText(f'Saving frame {frameIdx}...')
      ret, frame = cap.read()
      if not ret:
        print(f"Error reading at frame {frameIdx}")
        break
      frameHeight, frameWidth = frame.shape[:2]
      if frameHeight < height or frameWidth < width:
        xOffset = (width - frameWidth) // 2
        yOffset = (height - frameHeight) // 2
        copy = cv2.resize(frame, (width, height))
        copy[:] = 0
        copy[yOffset:yOffset+frameHeight, xOffset:xOffset+frameWidth] = frame
        frame = copy
      out.write(frame)
      if progressDialog.wasCanceled():
        cancelled = True
        break
    progressDialog.setLabelText('Saving video...')
    cap.release()
    out.release()
    progressDialog.close()


def chooseVideoToTroubleshootSplitVideo(self, controller):

  # Choosing video to split
  self.videoToTroubleshootSplitVideo, _ =  QFileDialog.getOpenFileName(self.window, "Select video", os.path.expanduser("~"), "All files(*)")
  if not self.videoToTroubleshootSplitVideo:
    return

  def beginningAndEndChosen():
    # Choosing directory to save sub-video
    firstFrame = self.configFile["firstFrame"]
    lastFrame = self.configFile["lastFrame"]
    self.configFile = configFile
    _, existingExtension = os.path.splitext(self.videoToTroubleshootSplitVideo)
    specialFormat = existingExtension in ('.sqb', '.seq', '.tif', '.tiff', '.bias')
    if specialFormat:
      existingExtension = '.avi'
    filename, _ = QFileDialog.getSaveFileName(controller.window, 'Choose where you want to save the sub-video.', os.path.join(os.path.dirname(self.videoToTroubleshootSplitVideo), f'subvideo{existingExtension}'), f"Video (*{existingExtension})")
    if not filename:
      return

    # Extracting sub-video
    cancelled = False
    progressDialog = QProgressDialog("Saving video...", "Cancel", firstFrame if specialFormat else 0, lastFrame if specialFormat else 100, controller.window, Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint)
    progressDialog.setWindowTitle('Saving Video')
    progressDialog.setAutoClose(False)
    progressDialog.setAutoReset(False)
    progressDialog.setWindowModality(Qt.WindowModality.ApplicationModal)
    progressDialog.setMinimumDuration(500)

    if specialFormat:
        _runLegacy(self.videoToTroubleshootSplitVideo, filename, progressDialog, firstFrame, lastFrame)
    else:
      numerator, denominator = map(float, subprocess.check_output(["ffprobe", "-v", "error", "-select_streams", "v:0", "-of", "default=noprint_wrappers=1:nokey=1", "-show_entries", "stream=r_frame_rate", self.videoToTroubleshootSplitVideo], universal_newlines=True).split('/'))
      fps = numerator / denominator
      cmd = ffmpeg_progress_yield.FfmpegProgress(["ffmpeg", "-y", "-ss", f"{firstFrame / fps}", "-to", f"{lastFrame / fps}", "-i", self.videoToTroubleshootSplitVideo, "-force_key_frames", "00:00:00", "-c", "copy", filename])
      for pct in cmd.run_command_with_progress():
          if progressDialog.wasCanceled():
            cancelled = True
            cmd.quit_gracefully()
            break
          progressDialog.setValue(int(pct))
      progressDialog.close()

    if not cancelled:
      self.show_frame("VideoToTroubleshootSplitVideo")
    else:
      self.show_frame("ChooseVideoToTroubleshootSplitVideo")

  configFile = self.configFile.copy()
  # User input of beginning and end of subvideo
  util.chooseBeginningPage(self, self.videoToTroubleshootSplitVideo, "Choose where the beginning of your sub-video should be", "Ok, I want the sub-video to start at this frame!",
                           lambda: util.chooseEndPage(self, self.videoToTroubleshootSplitVideo, "Choose where the sub-video should end.", "Ok, I want the sub-video to end at this frame!", beginningAndEndChosen,
                                                      leftButtonInfo=("Go to the start page", lambda: setattr(self, 'configFile', configFile) or self.show_frame("StartPage"))),
                           leftButtonInfo=("Go to the start page", lambda: self.show_frame("StartPage")))



from scipy.optimize import linear_sum_assignment
import zebrazoom.dataAPI as dataAPI
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd
import numpy as np
import shutil
import math
import h5py
import os


def detectBouts(videoName: str, nbWells: int, nbAnimalsPerWell: int, self):

  for numWell in range(nbWells):
    numAnimal = 0
    dataForEachAnimal = []
    for numAnimal in range(nbAnimalsPerWell):
      data = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, None, None, "HeadPos")
      dataForEachAnimal.append(data)
  
    for numAnimal in range(nbAnimalsPerWell):
      
      # data = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, None, None, "HeadPos")
      dataForAnimal = dataForEachAnimal[numAnimal]
    
      # for numAnimal, dataForAnimal in enumerate(data):
        
      frameDistance = self._hyperparameters["frameGapComparision"]
      auDessus = [int(math.sqrt(sum((dataForAnimal[i+frameDistance] - coords) ** 2)) >= self._hyperparameters["coordinatesOnlyBoutDetectionMinDistDataAPI"]) for i, coords in enumerate(dataForAnimal[:-frameDistance])]
      auDessus.extend([0] * frameDistance)
      
      auDessus2 = np.copy(auDessus)
      windowGap = self._hyperparameters["fillGapFrameNb"]
      if windowGap:
        for i in range(windowGap,nbFrames-windowGap):
          if auDessus[i] == 0:
            j = i - windowGap
            while (auDessus[j] == 0) and (j < i + windowGap):
              j = j + 1
            if j < i + windowGap:
              j = j + 1;
              while (auDessus[j] == 0) and (j < i + windowGap):
                j = j + 1;
              if auDessus[j] > 0:
                for k in range(i - windowGap, i + windowGap):
                  auDessus2[k] = 1
    
    numWell = 0
    numAnimal = 0
    with h5py.File(self._hyperparameters['H5filename'], 'a') as results:
      listOfBouts = results.require_group(f"dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts")
      inBout = 0
      nbFrames = len(dataForAnimal)
      boutIdx = 0
      for i in range(0, nbFrames):
        if (inBout == 0) and (auDessus2[i] == 1):
          firstFrame = i
          inBout = 1
        if (inBout == 1) and ((auDessus2[i] == 0) or (i == nbFrames-1)):
          if firstFrame != (i - 1):
            listOfBouts.attrs['numberOfBouts'] = boutIdx + 1
            boutGroup = listOfBouts.require_group(f'bout{boutIdx}')
            boutGroup.attrs["BoutStart"] = self._hyperparameters["firstFrame"] + firstFrame
            boutGroup.attrs["BoutEnd"]   = self._hyperparameters["firstFrame"] + i - 1
            boutIdx += 1
          inBout = 0



from scipy.optimize import linear_sum_assignment
import zebrazoom.dataAPI as dataAPI
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd
import numpy as np
import shutil
import math
import h5py
import os

def rolling_mean(data, window_size):
    # Ensure window size is odd for symmetric padding
    if window_size % 2 == 0:
        raise ValueError("Window size must be odd for symmetric padding.")
    
    half_window = window_size // 2
    
    # Apply rolling mean separately on each column
    cumsum = np.cumsum(np.insert(data, 0, 0, axis=0), axis=0)
    smoothed_data = (cumsum[window_size:] - cumsum[:-window_size]) / window_size
    
    # Pad the result to match the input size
    start_pad = np.tile(smoothed_data[0], (half_window, 1))
    end_pad = np.tile(smoothed_data[-1], (half_window, 1))
    smoothed_data = np.vstack((start_pad, smoothed_data, end_pad))
    
    return smoothed_data
def smoothHeadPositions(videoName: str, nbWells: int, nbAnimalsPerWell: int, smoothHeadPositionsWindow: int):

  for numWell in range(nbWells):
    dataForEachAnimal = []
    for numAnimal in range(nbAnimalsPerWell):
      data = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, None, None, "HeadPos")
      smoothed_data = rolling_mean(data, smoothHeadPositionsWindow)
      dataAPI.setDataPerTimeInterval(videoName, numWell, numAnimal, None, None, "HeadPos", smoothed_data)
